/*
 *  Harald Ruda \u00a9 2009, GPL applies
 *
 *  Code for School Master Clock
 *
 *  Control synchronization of clocks as well as two bell zones
 *
 */

//=====  Libraries Used  =====//

#include <PString.h>
#include <Streaming.h>
#include <Flash.h>
#include <DateTime.h>
#include <DateTimeStrings.h>
#include <DaylightSavings.h>
#include <LiquidCrystal.h>

//=====  Constants and Variables  =====//

#include "WProgram.h"
void setup();
void loop();
void timeCheck();
void dstCheck();
void displayTime();
void displayMessage();
void powerCheck();
void serialCheck();
void getVersion();
void getStatus();
void automaticDriftAdjustment(boolean v);
void automaticDaylightSavings(boolean v);
void getBellDuration();
void setBellDuration();
long readSerialValue();
void getDriftAdjustment();
void setDriftAdjustment();
void getUtime();
void getTime();
void setTime();
int setAlarms();
void ringBells();
void syncHour();
void syncDay();
void sendTime();
void sendTimePointer(time_t * p);
void sendTimeDigits(byte digits);
void bellNext(byte number);
void syncNext(byte number);
void bellCheck();
void syncCheck();
unsigned long constantBellDurationMilli = 3500;

byte statusLEDpin = 13;

byte batGreenLED = 8;
byte batRedLED = 9;

byte syncPin  = 12;
byte zone0Pin = 11;
byte zone1Pin = 10;

boolean timeHasBeenSet = false;
boolean adjustDaylightSavings = true;
boolean autoDriftAdjustment = true;

#define  STATUS_BLINK_INTERVAL  500
#define  POWER_CHECK_INTERVAL  1500

int batteryValue = 1023;
int dcValue = 1023;

#define  MAX_ALARMS  40

byte numAlarms = 0;
byte alarmHour[MAX_ALARMS];
byte alarmMinute[MAX_ALARMS];
//byte alarmSecond[MAX_ALARMS];
byte alarmZone[MAX_ALARMS];

time_t  bellNextTime = 0;
time_t  syncNextTime = 0;

boolean dst = false;

// initialize the library with the numbers of the interface pins
LiquidCrystal lcd(6, 7, 2, 3, 4, 5);

boolean bellRinging = false;
boolean clockSyncing = false;

char dateBuffer[17];
PString dateMessage(dateBuffer, sizeof(dateBuffer));
char bellBuffer[17];
PString bellMessage(bellBuffer, sizeof(bellBuffer));
char otherBuffer[17];
PString otherMessage(otherBuffer, sizeof(otherBuffer));
char smallBuffer[10];
PString smallString(smallBuffer, sizeof(smallBuffer));

char SPACE [2]  = " ";
char COLON [2]  = ":";
char ZERO  [2]  = "0";

//=====  Sketch Routines  =====//

//
//  Do this once
//
void setup()
{
  pinMode(statusLEDpin, OUTPUT);
  
  pinMode(batRedLED, OUTPUT);
  pinMode(batGreenLED, OUTPUT);
  
  pinMode(syncPin, OUTPUT);
  pinMode(zone0Pin, OUTPUT);
  pinMode(zone1Pin, OUTPUT);
  
  digitalWrite(syncPin, LOW);
  digitalWrite(zone0Pin, LOW);
  digitalWrite(zone1Pin, LOW);
  
  lcd.begin(16, 2);  // set up the LCD's number of rows and columns: 
  lcd.setCursor(0, 0);
  lcd << F("Ready!");
  
  Serial.begin(9600);
  Serial << F("Hello there, I'm ready!") << endl;
}

//
//  Do this over and over
//
void loop()
{
  timeCheck();       //  Check if time has been set, adjust status LED
  dstCheck();        //  Adjust daylight savings time, if necessary
  
  displayTime();     // Display the time on the LCD
  displayMessage();  // Display other messages on the LCD
  
  syncCheck();       //  Synchronize clocks
  bellCheck();       //  Ring the bells if it is time
  
  powerCheck();      //  Check main and battery voltages, adjust power LEDs
  
  serialCheck();     //  Deal with any input from the serial connection
  
  delay(100);        //  Wait 100 ms
}

//=====  Main Routines  =====//

//
//  Status LED (pin 13) will blink until the time has been set
//  After the time has been set, it will be off
//
void timeCheck()
{
  static long lastTime = 0;
  static boolean status = LOW;
  
  if (timeHasBeenSet && (status == LOW)) 
    return;
  
  if ((millis() - lastTime) < STATUS_BLINK_INTERVAL) 
    return;
  
  lastTime = millis();
  status = (status == LOW) ? HIGH : LOW;
  digitalWrite(statusLEDpin, status);
}


void dstCheck()
{
  static long lastTime = 0;
  
  if (!DateTime.available()) 
    return;
  
  if (!adjustDaylightSavings) 
    return;
  
  if ((millis() - lastTime) < 1000)  // Check once a second
    return;
  
  lastTime = millis();
  dst = DaylightSavings.check();
}


void displayTime()
{
  static time_t lastTime = 0;
  static byte lastDay = 32;
  
  if (DateTime.available()) 
  {
    time_t t = DateTime.now();
    
    if (t == lastTime) 
      return;
    
    lastTime = t;
    
    boolean am = true;
    byte hour, minute, second;
    byte day, dow, month, year;
    
    DateTime.localTime(&t, &second, &minute, &hour, &day, &dow, &month, &year);
    
    if (hour >= 12) 
        am = false;
    
    if (hour > 12) 
        hour -= 12;
    
    if (hour == 0) 
        hour = 12;
    
    lcd.setCursor(0, 0);
    lcd << ((hour < 10) ? SPACE : "") << _DEC(hour) << COLON;
    lcd << ((minute < 10) ? ZERO : "") << _DEC(minute);
    lcd << COLON <<  ((second < 10) ? ZERO : "") << _DEC(second);
    lcd << ((am) ? F(" AM") : F(" PM"));
    lcd << ((dst) ? F("  DST") : F("  STD"));
    
    if (day != lastDay) 
    {
      char * buf;
      
      lastDay = day;
      dateMessage.begin();
      
      buf = DateTimeStrings.dayStr(dow);
      buf[3] = 0;
      dateMessage << buf << F(", ");
      dateMessage << ((day < 10) ? SPACE : "") << _DEC(day);
      buf = DateTimeStrings.monthStr(month);
      buf[3] = 0;
      dateMessage << SPACE << buf << SPACE << _DEC(year + 1900);
    }
  }
  else if (lastTime == 0) 
  {
    lcd.setCursor(0, 0);
    lcd << F("Please set time!");
    lastTime = 1;
  }
}

/*
char * int2string(int v, char leading, int places)
{
  static char buffer[6] = "";
  int p, i, d;
  
  for (i = 0, d = 10000; d > 0; i++, d /= 10) 
  {
    p = v / d;
    buffer[i] = (p == 0) ? leading : '0' + p;
    if (p > 0) 
    {
      v -= p * d;
      leading = '0'; // no longer leading
    }
  }
  buffer[5] = 0;
  
  return &buffer[5 - places];
}
*/

void displayMessage()
{
  static long lastTime = 0;
  static byte status = 0;
  
  if (bellRinging) 
  {
    otherMessage.begin();
    otherMessage << F("BELL ringing NOW");
    lcd.setCursor(0, 1);
    lcd << otherBuffer;
    
    return;
  }
  
  if (clockSyncing) 
  {
    otherMessage.begin();
    otherMessage << F("CLOCKS sync NOW ");
    lcd.setCursor(0, 1);
    lcd << otherBuffer;
    
    return;
  }
  
  if ((millis() - lastTime) < 3000) // 3 seconds
    return;
  
  lcd.setCursor(0, 1);
  if (status == 0)  lcd << dateBuffer;
  if (status == 1)  lcd << bellBuffer;
  
  status += 1;
  status = status % 2;
  lastTime = millis();
}


//
//  Check the DC power in and the battery voltage
//
//  Green - everything is fine
//  Amber - replace battery soon
//  Red   - replace battery now
//
//  Fast blinking red - battery not connected
//  Blining red - replace it now!
//
//  Strobing - no DC power
//
void powerCheck()
{
  boolean power;
  int red, green;
  boolean change = false;
  
  static unsigned long lastPowerCheck = 0;
  static unsigned long lastBlinkTime = 0;
  static int currentBlinkValue = LOW;
  
  static int batteryGood = 750;     // 8.5 V
  static int batteryReplace = 660;  // 7.5 V
  static int batteryBad = 615;      // 7.0 V
  static int batteryConnected = 100;
  
  if ((millis() - lastPowerCheck) >= POWER_CHECK_INTERVAL) 
  {
    lastPowerCheck = millis();
    dcValue = analogRead(3);
    batteryValue = analogRead(1);
    change = true;
  }
  
  power = (dcValue > 100);
  red = (batteryValue < batteryGood) ? HIGH : LOW;
  green = (batteryValue > batteryReplace) ? HIGH : LOW;
  
  if (power && (batteryValue < batteryBad))  //  Blink quickly
  {
    long interval = (batteryValue < batteryConnected) ? 150 : 600;
    
    if (millis() - lastBlinkTime > interval) 
    {
      lastBlinkTime = millis();
      currentBlinkValue = (currentBlinkValue == LOW) ? HIGH : LOW;
      change = true;
    }
    
    if (currentBlinkValue == LOW) 
    {
      red = LOW;
      green = LOW;
    }
  }
  else if (!power)  //  Strobe every second
  {
    if (millis() - lastBlinkTime > 1000) 
    {
      lastBlinkTime = millis();
      digitalWrite(batRedLED, red);
      digitalWrite(batGreenLED, green);
      delay(5);
      change = true;
    }
    
    red = LOW;
    green = LOW;
  }
  
  if (change) 
  {
    digitalWrite(batRedLED, red);
    digitalWrite(batGreenLED, green);
  }
}


//
//  Process any input from the serial connection
//
void serialCheck()
{
  byte inByte;

  if (Serial.available() > 0) 
  {
    inByte = Serial.read();
    
    switch (inByte) 
    {
      case 'g': getTime(); break;
      case 's': setTime(); break;
      case 'a': setAlarms(); break;
      case 'r': ringBells(); break;
      case 'h': syncHour(); break;
      case 'd': syncDay(); break;
      case 'u': getUtime(); break;
      case 'v': getVersion(); break;
      case 't': getStatus(); break;
      
      case 'n': getBellDuration(); break;
      case 'm': setBellDuration(); break;
      case 'k': getDriftAdjustment(); break;
      case 'j': setDriftAdjustment(); break;
      case 'q': automaticDriftAdjustment(true); break;
      case 'w': automaticDriftAdjustment(false); break;
      case 'x': automaticDaylightSavings(true); break;
      case 'y': automaticDaylightSavings(false); break;
      
      default:
        Serial << F("Unknown command: 0x") << _HEX(inByte) << " ('" << _BYTE(inByte) << "')" << endl;
        break;
    }
  }
}


void getVersion()
{
  Serial << F("Version 1.0 2009-09-15") << endl;
}


void getStatus()
{
  Serial << F("Status:") << endl;
  Serial << "  dc = " << dcValue << endl;
  Serial << "  battery = " << batteryValue << endl;
  Serial << "  ";
  getTime();
  Serial << "  " << _DEC(numAlarms) << " alarms have been set." << endl;
  
  time_t t = DateTime.nowUnadjusted();
  Serial << F("  The unadjusted time is: ");
  sendTimePointer(&t);
  
  Serial << F("") << _DEC(DateTime.getDriftTime()) << endl;
  
  t = DaylightSavings.getSpringForward();
  Serial << F("  The spring-forward time is: ");
  sendTimePointer(&t);
  t = DaylightSavings.getFallBack();
  Serial << F("  The fall-back time is: ");
  sendTimePointer(&t);
}


void automaticDriftAdjustment(boolean v)
{
  autoDriftAdjustment = v;
}


void automaticDaylightSavings(boolean v)
{
  adjustDaylightSavings = v;
  
  if (!adjustDaylightSavings) 
  {
    dst = false;
    DateTime.setDSToffset(0);
  }
}


void getBellDuration()
{
  Serial << F("The bells currently ring for ") << _DEC(constantBellDurationMilli) << F(" ms") << endl;
}


void setBellDuration()
{
  long newDuration = readSerialValue();
  
  if (0 <= newDuration && newDuration < 60000) 
    constantBellDurationMilli = newDuration;
}


long readSerialValue()
{
  boolean negative = false;
  long value = 0;
  char c;
  
  while ((c = Serial.read()) != '.')
  {
    if (c == '-' && value == 0) 
      negative = true;
    else if (c >= '0' && c <= '9') 
      value = (10 * value) + (c - '0');
    else 
      Serial << F("Illegal character in message") << endl;   
  }
  
  return (negative) ? - value : value;
}


void getDriftAdjustment()
{
  Serial << F("The  current drift adjustment is ") << _DEC(DateTime.getDriftTime()) << F(" sec/day") << endl;
}


void setDriftAdjustment()
{
  long seconds = readSerialValue();
  
  Serial << F("Just read in: ") << _DEC(seconds) << endl;
  
  DateTime.setDriftTime(seconds);
}


void getUtime()
{
  unsigned long t = DateTime.now();
  
  Serial << _DEC(t) << endl;
}


void getTime()
{
  if (DateTime.available()) 
  {
    time_t t = DateTime.now();
    Serial << F("The current time is: ");
    sendTimePointer(&t);
  }
  else 
  {
    Serial << F("Time has not been set.") << endl;
  }
}


void setTime()
{
  static time_t lastSync = 0;
  
  if (Serial.available() >= 10) 
  {
    time_t theTime = 0;
    
    for (int i = 0; i < 10; i++)
    {
      char c = Serial.read();
      
      if (c >= '0' && c <= '9') 
        theTime = (10 * theTime) + (c - '0');
      else 
      {
        Serial << F("Illegal character in time sync message") << endl;
      }
    }
    
    Serial << F("last-sync ") << _DEC(lastSync) << endl;
    
    if (lastSync != 0) 
    {
      time_t since = theTime - lastSync;
      time_t drift = theTime - DateTime.nowRaw();
      int seconds = drift * 86000 / since;
        
      Serial << F("since ") << _DEC(since) << endl;
      Serial << F("drift ") << _DEC(drift) << endl;
      Serial << F("seconds ") << _DEC(seconds) << endl;
    }
    
    DateTime.sync(theTime, autoDriftAdjustment);
    timeHasBeenSet = true;
    Serial << F("The time has been set") << endl;
    
    syncNext(0);
    lastSync = theTime;
  }
  else 
  {
    Serial << F("Not enough characters in time sync message.") << endl;
  }
}


int setAlarms()
{
  byte newNumAlarms = 0;
  
  Serial << F("command: Set Alarms") << endl;
  
  if (!Serial.available()) 
    return 1;
  
  newNumAlarms = Serial.read();
  numAlarms = 0;  // Set to zero in case something goes wrong
  Serial << F("setting ") << _DEC(newNumAlarms) << F(" alarms.") << endl;
  
  for (int alarm = 0; alarm < newNumAlarms; alarm++) 
  {
    if (!Serial.available()) 
      return 1;
    
    alarmHour[alarm] = Serial.read();
    
    if (!Serial.available()) 
      return 1;
    
    alarmMinute[alarm] = Serial.read();
    
//    if (!Serial.available()) 
//      return 1;
    
//    alarmSecond[alarm] = Serial.read();
    
    if (!Serial.available()) 
      return 1;
    
    alarmZone[alarm] = Serial.read();
    
//    if (!Serial.available()) 
//      return 1;
    
//    alarmDuration[alarm] =  100 * Serial.read();
    
    Serial.print("Alarm ");
    Serial.print(alarm);
    Serial.print(" set for ");
    Serial.print(alarmHour[alarm], DEC);
    sendTimeDigits(alarmMinute[alarm]);
//    sendTimeDigits(alarmSecond[alarm]);
    Serial.print(" in zone:");
    if (bitRead(alarmZone[alarm], 0) == 1) 
        Serial.print(" Indoor");
    if (bitRead(alarmZone[alarm], 1) == 1) 
        Serial.print(" Outdoor");
    Serial.println();
//    Serial.print(", with a duration of ");
//    Serial.print(alarmDuration[alarm], DEC);
//    Serial.println(" ms");
  }
  
  numAlarms = newNumAlarms;  //  Safe to set now
  bellNext(0);
  
  Serial << F("Alarms set successfully!") << endl;
  
  return 0;
}


boolean bellActive = false;
boolean syncActive = false;

byte bellNumber;
byte syncNumber;

byte bellZone = 0;

unsigned long bellDurationSec;
unsigned long bellDurationMilli;

unsigned long syncDurationSec;
unsigned long syncDurationMilli;

unsigned long syncHourLength = 25 * 1000UL; // ms
unsigned long sync12HourLength = 25 * 60 * 1000UL; // ms


void ringBells()
{
  Serial << F("command: Ring Bells") << endl;
  
  if (bellActive) 
    return;
  
  if (numAlarms == 0) 
      numAlarms = 1;
      
  bellNumber--;
  bellNextTime = DateTime.now();
  bellZone = 0xFF;
  bellDurationMilli = 1500;
  bellDurationSec = bellDurationMilli / 1000;
}

void syncHour()
{
  Serial << F("command: Send Hourly Sync") << endl;
  
  if (syncActive) 
    return;
  
  syncNumber--;
  syncNextTime = DateTime.now();
  syncDurationMilli = syncHourLength;
  syncDurationSec = syncDurationMilli / 1000;
}

void syncDay()
{
  
  if (syncActive) 
    return;
  
  syncNumber--;
  syncNextTime = DateTime.now();
  syncDurationMilli = sync12HourLength;
  syncDurationSec = syncDurationMilli / 1000;
}


void sendTime()
{
  Serial.print(DateTime.Hour, DEC);  
  sendTimeDigits(DateTime.Minute);  
  sendTimeDigits(DateTime.Second); 
  Serial.print(" "); 
  Serial.print(DateTimeStrings.dayStr(DateTime.DayofWeek));
  Serial.print(", ");  
  Serial.print(DateTimeStrings.monthStr(DateTime.Month));  
  Serial.print(" ");
  Serial.print(DateTime.Day, DEC);   
  Serial.print(", ");
  Serial.println(DateTime.Year + 1900, DEC);   
}

void sendTimePointer(time_t * p)
{
  byte hour;
  byte minute;
  byte second;
  byte day;
  byte dow;
  byte month;
  byte year;
  
  DateTime.localTime(p, &second, &minute, &hour, &day, &dow, &month, &year);
  
  Serial.print(hour, DEC);  
  sendTimeDigits(minute);  
  sendTimeDigits(second); 
  Serial.print(" "); 
  Serial.print(DateTimeStrings.dayStr(dow));
  Serial.print(", ");  
  Serial.print(DateTimeStrings.monthStr(month));  
  Serial.print(" ");
  Serial.print(day, DEC);   
  Serial.print(", ");
  Serial.println(year + 1900, DEC);   
}

void sendTimeDigits(byte digits)
{
  Serial.print(":");
  if (digits < 10)
    Serial.print('0');
  Serial.print(digits, DEC);   
}

unsigned long bellMargin = 5;
unsigned long bellStart;

unsigned long syncMargin = 5;
unsigned long syncStart;


void bellNext(byte number)
{
  if (!DateTime.available()) 
    return;
  
  if (numAlarms == 0) 
    return;
  
  byte day = DateTime.Day;
  
  if (DateTime.DayofWeek == 0) // Sunday
    day += 1;
  
  if (DateTime.DayofWeek == 6) // Saturday
    day += 2;
 
  if (number >= numAlarms) 
  {
    number = 0;
    if (day == DateTime.Day) 
        day++;
  }
  
  bellNextTime = DateTime.makeTime(0, alarmMinute[number], alarmHour[number], day, DateTime.Month, DateTime.Year + 1900);
//  bellNextTime = DateTime.makeTime(alarmSecond[number], alarmMinute[number], alarmHour[number], day, DateTime.Month, DateTime.Year + 1900);
  
  bellZone = alarmZone[number];
  
  bellDurationMilli = constantBellDurationMilli;
//    bellDurationMilli = alarmDuration[number];
  bellDurationSec = bellDurationMilli / 1000;
  
  bellNumber = number;
  
  Serial.print("The bell will ring at: ");
  sendTimePointer(&bellNextTime);
  
  {
    boolean am = true;
    byte hour, minute;
    
    DateTime.localTime(&bellNextTime, NULL, &minute, &hour, NULL, NULL, NULL, NULL);
    
    if (hour >= 12) 
        am = false;
      
    if (hour > 12) 
        hour -= 12;
    
    if (hour == 0) 
        hour = 12;
    
    bellMessage.begin();
    bellMessage << F(" Bell @ ");
    bellMessage << ((hour < 10) ? SPACE : "") << _DEC(hour) << COLON;
    bellMessage << ((minute < 10) ? ZERO : "") << _DEC(minute);
    bellMessage << ((am) ? F(" AM") : F(" PM"));
  }
}

byte syncMinute = 00;
byte syncSecond = 00;

void syncNext(byte number)
{
  if (!DateTime.available()) 
    return;
  
  byte day = DateTime.Day;
  
  if (number >= 24) 
  {
    day++;
    number = 0;
  }
  
  syncDurationMilli = (number % 12 == 6) ? sync12HourLength : syncHourLength;
  syncDurationSec = syncDurationMilli / 1000;
  
  syncNextTime = DateTime.makeTime(syncSecond, syncMinute, number, day, DateTime.Month, DateTime.Year + 1900);
  
  syncNumber = number;
  
  Serial << F("The clocks will sync at: ");
  sendTimePointer(&syncNextTime);
}

void bellCheck()
{
  time_t currentTime = DateTime.now();
  
  if (numAlarms > 0 && currentTime >= bellNextTime) 
  {
    if (currentTime > bellNextTime + bellDurationSec + bellMargin) // Catching up 
    {
      Serial << F("Catching up ... ");
      bellNext(bellNumber + 1);
    }
    else if (bellActive) // Previously turned on
    {
      if (millis() - bellStart >= bellDurationMilli)  // Time to turn off
      {
        bellRinging = false;
        bellActive = false;
        if (bitRead(bellZone, 0) == 1) 
            digitalWrite(zone0Pin, LOW);
        if (bitRead(bellZone, 1) == 1) 
            digitalWrite(zone1Pin, LOW);
        Serial << F("Bell OFF at ");
        sendTimePointer(&currentTime);
        bellNext(bellNumber + 1); // Schedule next bell
      }
    }
    else // Turn on now
    {
      bellRinging = true;
      bellActive = true;
      bellStart = millis();
      if (bitRead(bellZone, 0) == 1) 
          digitalWrite(zone0Pin, HIGH);
      if (bitRead(bellZone, 1) == 1) 
          digitalWrite(zone1Pin, HIGH);
      Serial << F("Bell ON at ");
      sendTimePointer(&currentTime);
    }
  }
}


void syncCheck()
{
  time_t currentTime = DateTime.now();
  
  if (DateTime.available() && currentTime >= syncNextTime) 
  {
    if (currentTime > syncNextTime + syncDurationSec + syncMargin)  // Catching up 
    {
      Serial << F("Catching up ... ");
      syncNext(syncNumber + 1);
    }
    else if (syncActive) // Previously turned on
    {
      if (millis() - syncStart >= syncDurationMilli)  // Time to turn off
      {
        clockSyncing = false;
        syncActive = false;
        digitalWrite(syncPin, LOW);
        Serial << F("Sync OFF at ");
        sendTimePointer(&currentTime);
        syncNext(syncNumber + 1);  // Schedule next sync
      }
    }
    else // Turn on now
    {
      clockSyncing = true;
      syncActive = true;
      syncStart = millis();
      digitalWrite(syncPin, HIGH);
      Serial << F("Sync ON at ");
      sendTimePointer(&currentTime);
    }
  }
}



int main(void)
{
	init();

	setup();
    
	for (;;)
		loop();
        
	return 0;
}

